#include<math.h>
#include<stdio.h>
#include<assert.h>
#include<gsl/gsl_interp.h>
#include<gsl/gsl_spline.h>
#include<gsl/gsl_vector.h>
#include<gsl/gsl_matrix.h>
#include<gsl/gsl_blas.h>
#include<gsl/gsl_linalg.h>

typedef struct {double* p; double* c;} quad_coefs;
typedef struct {double* b; double* c; double* d;} cubic_coefs;

int search(int n, double* x, double z){
	assert(x[0]<=z && z<=x[n-1]);
	int i=0, j=n-1;
	while(j-i>1){
		int mid=(i+j)/2;
		if(z>x[mid]) i=mid; else j=mid;
	}
return i;
}

double lin_interpol(int n, double x[], double y[], double z) {
	int i=search(n,x,z);
	double pi = (y[i+1]-y[i])/(x[i+1]-x[i]);
return y[i]+pi*(z-x[i]);
}

double lin_interpol_integ(int n, double x[], double y[], double z) {
	int i=search(n,x,z);
	double sum = 0;
	for(int j=0;j<i;j++) {
		double pi = (y[j+1]-y[j])/(x[j+1]-x[j]);
		sum += y[j]*x[j+1]+pi*(pow(x[j+1],2)/2-x[j]*x[j+1])-y[j]*x[j]+pi*pow(x[j],2)/2;
	}
	double pi = (y[i+1]-y[i])/(x[i+1]-x[i]);
	sum += y[i]*z+pi*(pow(z,2)/2-x[i]*z)-y[i]*x[i]+pi*pow(x[i],2)/2;
return sum;
}

void quad_interpol(int n, double x[], double y[], quad_coefs* a) {
	for(int i=0;i<(n-1);i++) {
		(*a).p[i] = (y[i+1]-y[i])/(x[i+1]-x[i]);
	}
	double c_n = 0;
	for(int i=0;i<(n-2);i++) {
		c_n = ((*a).p[i+1]-(*a).p[i]-c_n*(x[i+1]-x[i]))/(x[i+2]-x[i+1]);
	}
	(*a).c[n-2] = c_n/2;
	for(int i=(n-3);i>=0;i--) {
		(*a).c[i] = ((*a).p[i+1]-(*a).p[i]-(*a).c[i+1]*(x[i+2]-x[i+1]))/(x[i+1]-x[i]);
	}
return;
}

double quad_eval(quad_coefs coefs, int n, double* x, double* y, double z) {
	int i = search(n,x,z);
	double b_i = (coefs.p)[i]-(coefs.c)[i]*(x[i+1]-x[i]);
return y[i]+b_i*(z-x[i])+(coefs.c)[i]*pow((z-x[i]),2);
}

double quad_diff_eval(quad_coefs coefs, int n, double* x, double* y, double z) {
	int i = search(n,x,z);
	double b_i = (coefs.p)[i]-(coefs.c)[i]*(x[i+1]-x[i]);
return b_i+2*(coefs.c)[i]*(z-x[i]);
}

double quad_integ_eval(quad_coefs coefs, int n, double* x, double* y, double z) {
	int i=search(n,x,z);
	double sum = 0;
	for(int j=0;j<i;j++) {
		double b_j = (coefs.p)[j]-(coefs.c)[j]*(x[j+1]-x[j]);
		sum += y[j]*(x[j+1]-x[j])+b_j/2*pow((x[j+1]-x[j]),2)+(coefs.c)[j]/3*pow((x[j+1]-x[j]),3);
	}
	double b_i = (coefs.p)[i]-(coefs.c)[i]*(x[i+1]-x[i]);
	sum += y[i]*(z-x[i])+b_i/2*pow((z-x[i]),2)+(coefs.c)[i]/3*pow((z-x[i]),3);
return sum;
}

void cubic_interpol(int n, double x[], double y[], cubic_coefs* a) {
	gsl_matrix* A = gsl_matrix_alloc(n,n);
	gsl_vector* b = gsl_vector_alloc(n);
	gsl_vector* B = gsl_vector_alloc(n);
	gsl_vector_set(B,0,(3*(y[1]-y[0])/(x[1]-x[0])));
	gsl_vector_set(B,n-1,(3*(y[n-1]-y[n-2])/(x[n-1]-x[n-2])));
	for(int i=0;i<n-2;i++) {
		double B_i1 = 3*((y[i+1]-y[i])/(x[i+1]-x[i])+(y[i+2]-y[i+1])/(x[i+2]-x[i+1])*(x[i+1]-x[i])/(x[i+2]-x[i+1]));
		gsl_vector_set(B,i+1,B_i1);
	}
	for(int i=0;i<n;i++) {
		for(int j=0;j<n;j++) {
			if(abs(i-j)>1) {
				gsl_matrix_set(A,i,j,0);
			}
			if((i-j)==1) {
				gsl_matrix_set(A,i,j,1);
			}
			if((i-j==-1)) {
				if(i==0) {gsl_matrix_set(A,i,j,1);}
				else {
					double Qi = (x[i]-x[i-1])/(x[i+1]-x[i]);
					gsl_matrix_set(A,i,j,Qi);
				}
			}
			if((i-j)==0) {
				if(i==0 || i==n-1) {gsl_matrix_set(A,i,j,2);}
				else {
					double Di = 2*((x[i]-x[i-1])/(x[i+1]-x[i]))+2;
					gsl_matrix_set(A,i,j,Di);
				}
			}
		}
	}
	gsl_linalg_HH_solve(A,B,b);
	for(int i=0;i<n-1;i++) {
		double pi = (y[i+1]-y[i])/(x[i+1]-x[i]);
		double bi = gsl_vector_get(b,i);
		double bi1 = gsl_vector_get(b,i+1);
		(*a).b[i] = bi;
		(*a).c[i] = (3*pi-2*bi-bi1)/(x[i+1]-x[i]);
		(*a).d[i] = (bi+bi1-2*pi)/pow((x[i+1]-x[i]),2);
	}
	gsl_matrix_free(A);
	gsl_vector_free(b);
	gsl_vector_free(B);
}

double cubic_eval(cubic_coefs coefs, int n, double* x, double* y, double z) {
	int i = search(n,x,z);
return y[i]+(coefs.b)[i]*(z-x[i])+(coefs.c)[i]*pow((z-x[i]),2)+(coefs.d)[i]*pow((z-x[i]),3);
}

double cubic_diff_eval(cubic_coefs coefs, int n, double* x, double* y, double z) {
	int i = search(n,x,z);
return (coefs.b)[i]+2*(coefs.c)[i]*(z-x[i])+3*(coefs.d)[i]*pow((z-x[i]),2);
}

double cubic_integ_eval(cubic_coefs coefs, int n, double* x, double* y, double z) {
	int i=search(n,x,z);
	double sum = 0;
	for(int j=0;j<i;j++) {
		sum += y[j]*(x[j+1]-x[j])+(coefs.b)[j]/2*pow((x[j+1]-x[j]),2)+\
		(coefs.c)[j]/3*pow((x[j+1]-x[j]),3)+(coefs.d)[j]/4*pow((x[j+1]-x[j]),4);
	}
	sum += y[i]*(z-x[i])+(coefs.b)[i]/2*pow((z-x[i]),2)+(coefs.c)[i]/3*pow((z-x[i]),3)+(coefs.d)[i]/4*pow((z-x[i]),4);
return sum;
}

double func(double x) {
return 25.-pow(x-5.,2.);
}
int main() {
	int n=6;
	double x[n];
	double y[n];
	for(int i=0;i<n;i++) {
		x[i]=2*i;
		y[i]=func((double)(2*i));
	}

	FILE* lin_out_stream=fopen("lin_out.txt","w");
	gsl_interp* linear_space = gsl_interp_alloc(gsl_interp_linear,n);
	gsl_interp_init(linear_space,x,y,n);
	for(double j=0.1;j<=10;j+=1./16) {
		double my_lin = lin_interpol(n,x,y,j);
		double gsl_lin = gsl_interp_eval(linear_space,x,y,j,NULL);
		double func_value = func(j);
		double my_lin_integ = lin_interpol_integ(n,x,y,j);
		double gsl_lin_integ = gsl_interp_eval_integ(linear_space,x,y,x[0],j,NULL);
		double integ_value = 5.0*pow(j,2)-pow(j,3)/3;
		fprintf(lin_out_stream,"%g %g %g %g %g %g %g\n",j,my_lin,gsl_lin,func_value,my_lin_integ,gsl_lin_integ,integ_value);
	}
	gsl_interp_free(linear_space);
	fclose(lin_out_stream);

	FILE* quad_out_stream=fopen("quad_out.txt","w");
	double my_quad_p[n-1];
	double my_quad_c[n-1];
	quad_coefs my_quad = {.p=my_quad_p,.c=my_quad_c};
	quad_interpol(n,x,y,&my_quad);
	gsl_interp* quad_space = gsl_interp_alloc(gsl_interp_polynomial,n);
	gsl_interp_init(quad_space,x,y,n);
	for(double j=0.1;j<=10;j+=1./16) {
		double my_quad_value = quad_eval(my_quad,n,x,y,j);
		double gsl_quad = gsl_interp_eval(quad_space,x,y,j,NULL);
		double func_value = func(j);
		double my_quad_diff = quad_diff_eval(my_quad,n,x,y,j);
		double gsl_quad_diff = gsl_interp_eval_deriv(quad_space,x,y,j,NULL);
		double diff_value = 2*(5-j);
		double my_quad_integ = quad_integ_eval(my_quad,n,x,y,j);
		double gsl_quad_integ = gsl_interp_eval_integ(quad_space,x,y,x[0],j,NULL);
		double integ_value = 5.0*pow(j,2)-pow(j,3)/3;
		fprintf(quad_out_stream,"%g %g %g %g %g %g %g %g %g %g\n",\
		j,my_quad_value,gsl_quad,func_value,my_quad_diff,gsl_quad_diff,diff_value,my_quad_integ,gsl_quad_integ,integ_value);
	}
	gsl_interp_free(quad_space);
	fclose(quad_out_stream);

	FILE* cubic_out_stream=fopen("cubic_out.txt","w");
	double my_cubic_b[n-1];
	double my_cubic_c[n-1];
	double my_cubic_d[n-1];
	cubic_coefs my_cubic = {.b=my_cubic_b,.c=my_cubic_c,.d=my_cubic_d};
	cubic_interpol(n,x,y,&my_cubic);
	gsl_interp* cubic_space = gsl_interp_alloc(gsl_interp_cspline,n);
	gsl_interp_init(cubic_space,x,y,n);
	for(double j=0.1;j<=10;j+=1./16) {
		double my_cubic_value = cubic_eval(my_cubic,n,x,y,j);
		double gsl_cubic = gsl_interp_eval(cubic_space,x,y,j,NULL);
		double func_value = func(j);
		double my_cubic_diff = cubic_diff_eval(my_cubic,n,x,y,j);
		double gsl_cubic_diff = gsl_interp_eval_deriv(cubic_space,x,y,j,NULL);
		double diff_value = 2*(5-j);
		double my_cubic_integ = cubic_integ_eval(my_cubic,n,x,y,j);
		double gsl_cubic_integ = gsl_interp_eval_integ(cubic_space,x,y,x[0],j,NULL);
		double integ_value = 5.0*pow(j,2)-pow(j,3)/3;
		fprintf(quad_out_stream,"%g %g %g %g %g %g %g %g %g %g\n",\
		j,my_cubic_value,gsl_cubic,func_value,my_cubic_diff,gsl_cubic_diff,diff_value,my_cubic_integ,gsl_cubic_integ,integ_value);
	}
	gsl_interp_free(cubic_space);
	fclose(cubic_out_stream);
return 0;
}