#include "functions.h"

void print_vector(char s[], gsl_vector* v) {
	printf("%s\n",s);
	for(int i=0;i<v->size;i++) {
		printf("%15g",gsl_vector_get(v,i));
		printf("\n");
	}
}

void print_matrix(FILE* stream,gsl_matrix* M) {
    for(int i = 0; i < M->size1; i++) {
        for(int j = 0; j < M->size2; j++) {
            fprintf(stream,"%15g\t", gsl_matrix_get(M, i, j));
        }
        fprintf(stream,"\n");
    }
}

void GS_decomp(gsl_matrix* A, gsl_matrix* R) {
    int n = A -> size1;
    int m = A -> size2;
    for(int i=0;i<m;i++) {
        gsl_vector_view ai = gsl_matrix_column(A,i);
        double Rii = gsl_blas_dnrm2(&(ai.vector));
        gsl_matrix_set(R,i,i,Rii);
        gsl_vector_scale(&(ai.vector),1/Rii);
        for(int j=i+1;j<m;j++) {
            gsl_vector_view aj = gsl_matrix_column(A,j);
            double Rij;
            gsl_blas_ddot(&(ai.vector),&(aj.vector),&Rij);
            gsl_matrix_set(R,i,j,Rij);
            gsl_vector* qiRij = gsl_vector_alloc(n);
            gsl_vector_memcpy(qiRij,&(ai.vector));
            gsl_vector_scale(qiRij,Rij);
            gsl_vector_sub(&(aj.vector),qiRij);
            gsl_vector_free(qiRij);
        }
    }
}

void back_sub(gsl_matrix* R, gsl_vector* x, gsl_vector* b) {
    int m = x->size;
    for(int i=m-1;i>=0;i--) {
        double xi = gsl_vector_get(b,i);
        for(int j=i+1;j<m;j++) {
            xi += -gsl_matrix_get(R,i,j)*gsl_vector_get(x,j);
        }
        xi /= gsl_matrix_get(R,i,i);
        gsl_vector_set(x,i,xi);
    }
}

void GS_solve(gsl_matrix* Q, gsl_matrix* R, gsl_vector* b, gsl_vector* x) {
    assert(Q->size1==Q->size2);
    gsl_blas_dgemv(CblasTrans,1,Q,b,0,x);
    back_sub(R,x,x);
}

void GS_inverse(gsl_matrix* Q, gsl_matrix* R, gsl_matrix* B) {
    for(int j=0;j<Q->size2;j++) {
        gsl_vector* ej = gsl_vector_alloc(Q->size1);
        gsl_vector_set_basis(ej,j);
        gsl_vector_view Bj = gsl_matrix_column(B,j);
        GS_solve(Q,R,ej,&(Bj.vector));
        gsl_vector_free(ej);
    }
}

void LS_fit(data data, double (*f)(int,double), gsl_vector* c, gsl_vector* dc, gsl_matrix* cova) {
    int nr_data = (data.xs)->size;
    int nr_params = c->size;
    gsl_matrix* A = gsl_matrix_alloc(nr_data,nr_params);
    gsl_matrix* AtA = gsl_matrix_alloc(nr_params,nr_params);
    gsl_matrix* R = gsl_matrix_alloc(nr_params,nr_params);
    gsl_vector* b = gsl_vector_alloc(nr_data);
    for(int i=0; i<nr_data; i++) {
        for(int k=0; k<nr_params; k++) {
            double Aik = f(k,gsl_vector_get(data.xs,i))/gsl_vector_get(data.dy,i);
            gsl_matrix_set(A,i,k,Aik);
        }
        double bi = gsl_vector_get(data.ys,i)/gsl_vector_get(data.dy,i);
        gsl_vector_set(b,i,bi);
    }
    gsl_blas_dgemm(CblasTrans,CblasNoTrans,1,A,A,0,AtA);

    GS_decomp(A,R);
    gsl_blas_dgemv(CblasTrans,1,A,b,0,c);
    back_sub(R,c,c);
    GS_decomp(AtA,R);
    GS_inverse(AtA,R,cova);
    for(int i=0;i<nr_params;i++) {gsl_vector_set(dc,i,sqrt(gsl_matrix_get(cova,i,i)));}

    gsl_matrix_free(A);
    gsl_matrix_free(AtA);
    gsl_matrix_free(R);
    gsl_vector_free(b);
}